@@ -32,7 +32,8 @@ from ..piv import (
     PivController, ALGO, OBJ, SW, SLOT, PIN_POLICY, TOUCH_POLICY,
     DEFAULT_MANAGEMENT_KEY)
 from ..driver_ccid import APDUError, SW_APPLICATION_NOT_FOUND
-from .util import click_skip_on_help, click_callback, prompt_for_touch
+from .util import (
+    click_force_option, click_skip_on_help, click_callback, prompt_for_touch)
 from cryptography import x509
 from cryptography.exceptions import InvalidSignature
 from cryptography.hazmat.primitives import hashes, serialization
@@ -125,6 +126,8 @@ def info(ctx):
         click.echo('PUK blocked.')
     if controller.has_derived_key:
         click.echo('Management key is derived from PIN.')
+    if controller.has_stored_key:
+        click.echo('Management key is stored on device and protected by PIN.')
     try:
         chuid = b2a_hex(controller.get_data(OBJ.CHUID)).decode()
     except APDUError as e:
@@ -207,14 +210,8 @@ def generate_key(
     PUBLIC-KEY  File containing the generated public key. Use '-' to use stdout.
     """
     controller = ctx.obj['controller']
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(ctx)
-        _authenticate(ctx, controller, management_key)
+
+    _ensure_authenticated(ctx, controller, pin, management_key)
 
     algorithm = ALGO.from_string(algorithm)
 
@@ -259,14 +256,7 @@ def import_certificate(
     CERTIFICATE     File containing the certificate. Use '-' to use stdin.
     """
     controller = ctx.obj['controller']
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(ctx)
-        _authenticate(ctx, controller, management_key)
+    _ensure_authenticated(ctx, controller, pin, management_key)
 
     data = cert.read()
 
@@ -314,14 +304,7 @@ def import_key(
     PRIVATE-KEY File containing the private key. Use '-' to use stdin.
     """
     controller = ctx.obj['controller']
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(ctx)
-        _authenticate(ctx, controller, management_key)
+    _ensure_authenticated(ctx, controller, pin, management_key)
 
     data = private_key.read()
 
@@ -415,14 +398,7 @@ def set_chuid(ctx, management_key, pin):
     Generate and set a CHUID on the device.
     """
     controller = ctx.obj['controller']
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(ctx)
-        _authenticate(ctx, controller, management_key)
+    _ensure_authenticated(ctx, controller, pin, management_key)
     controller.update_chuid()
 
 
@@ -435,14 +411,7 @@ def set_ccc(ctx, management_key, pin):
     Generate and set a CCC on the device.
     """
     controller = ctx.obj['controller']
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(ctx)
-        _authenticate(ctx, controller, management_key)
+    _ensure_authenticated(ctx, controller, pin, management_key)
     controller.update_ccc()
 
 
@@ -459,17 +428,8 @@ def set_pin_retries(ctx, management_key, pin, pin_retries, puk_retries):
     Set the number of PIN and PUK retries.
     """
     controller = ctx.obj['controller']
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(ctx)
-        _authenticate(ctx, controller, management_key)
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
+    _ensure_authenticated(
+        ctx, controller, pin, management_key, require_pin_and_key=True)
     try:
         controller.set_pin_retries(pin_retries, puk_retries)
     except:
@@ -502,18 +462,8 @@ def generate_certificate(
     PUBLIC-KEY      File containing a public key. Use '-' to use stdin.
     """
     controller = ctx.obj['controller']
-
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(ctx)
-        _authenticate(ctx, controller, management_key)
-        if not pin:
-            pin = _prompt_pin(ctx)
-        _verify_pin(ctx, controller, pin)
+    _ensure_authenticated(
+        ctx, controller, pin, management_key, require_pin_and_key=True)
 
     data = public_key.read()
     public_key = serialization.load_pem_public_key(
@@ -584,8 +534,6 @@ def generate_certificate_signing_request(
     CSR         File to write CSR to. Use '-' to use stdout.
     """
     controller = ctx.obj['controller']
-    if not pin:
-        pin = _prompt_pin(ctx)
     _verify_pin(ctx, controller, pin)
 
     data = public_key.read()
@@ -616,14 +564,7 @@ def delete_certificate(ctx, slot, management_key, pin):
     Delete a certificate from a slot on the device.
     """
     controller = ctx.obj['controller']
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(ctx)
-        _authenticate(ctx, controller, management_key)
+    _ensure_authenticated(ctx, controller, pin, management_key)
     controller.delete_certificate(slot)
 
 
@@ -685,56 +626,72 @@ def change_puk(ctx, puk, new_puk):
     '-t', '--touch', is_flag=True,
     help='Require touch on YubiKey when prompted for management key.')
 @click.option('-n', '--new-management-key', help='A new management key.')
-@click.option(
-    '-d', '--derive-from-pin', is_flag=True,
-    help='Derive the management key from the current PIN code. \
-            Blocks the PUK code.')
 @click.option(
     '-m', '--management-key', help='Current management key.',
     callback=click_parse_management_key)
+@click.option(
+    '-p', '--protect', is_flag=True,
+    help='Store new management key on device, protected by PIN.'
+         ' Will generate a random key if no key is provided.')
+@click_force_option
 def change_management_key(
-        ctx, management_key, pin, new_management_key, touch, derive_from_pin):
+        ctx, management_key, pin, new_management_key, touch, protect, force):
     """
     Change the management key.
 
     Management functionality is guarded by a 24 byte management key.
     This key is required for administrative tasks, such as generating key pairs.
+    A random key may be generated and stored on the device, protected by PIN.
     """
     controller = ctx.obj['controller']
 
-    if controller.has_derived_key:
-        if not pin:
-            pin = _prompt_pin(pin)
-        _verify_pin(ctx, controller, pin)
-    else:
-        if not management_key:
-            management_key = _prompt_management_key(
-                ctx, prompt='Enter your current management key'
-                            ' [blank to use the default key]')
-        _authenticate(ctx, controller, management_key)
+    _ensure_authenticated(
+        ctx, controller, pin, management_key,
+        require_pin_and_key=protect,
+        mgm_key_prompt='Enter your current management key '
+                       '[blank to use default key]')
 
     # Touch not supported on NEO.
     if touch and controller.version < (4, 0, 0):
         ctx.fail('Require touch not supported on your device.')
 
-    if derive_from_pin:
-        if not pin:
-            pin = _prompt_pin(pin)
-        controller.use_derived_key(pin, touch=touch)
-    else:
-        if not new_management_key:
+    # If an old stored key needs to be cleared, the PIN is needed.
+    if not protect and controller.has_stored_key:
+        if pin:
+            _verify_pin(ctx, controller, pin)
+        else:
+            force or click.confirm(
+                    'The current management key is stored on the device'
+                    ' and will not be cleared if no PIN is provided. Continue?',
+                    abort=True)
+
+    # If the key should be protected by PIN and no key is given,
+    # we generate a random key.
+    if protect and not new_management_key:
+        new_management_key = _generate_random_management_key()
+
+    if not new_management_key:
+        if not force:
             new_management_key = click.prompt(
-                'Enter your new management key',
+                'Enter your new management key'
+                ' [blank to randomly generate]',
                 default='', show_default=False,
                 hide_input=True, confirmation_prompt=True)
-        try:
-            new_management_key = a2b_hex(new_management_key)
-        except:
-            ctx.fail('New management key has the wrong format.')
-        try:
-            controller.set_mgm_key(new_management_key, touch=touch)
-        except APDUError:
-            ctx.fail('Changing the management key failed.')
+
+        if force or new_management_key == '':
+            new_management_key = _generate_random_management_key()
+            click.echo(
+                'Generated management key: {}'.format(new_management_key))
+
+    try:
+        new_management_key = a2b_hex(new_management_key)
+    except:
+        ctx.fail('New management key has the wrong format.')
+    try:
+        controller.set_mgm_key(
+            new_management_key, touch=touch, store_on_device=protect)
+    except APDUError:
+        ctx.fail('Changing the management key failed.')
 
 
 @piv.command('unblock-pin')
@@ -769,19 +726,46 @@ def _prompt_management_key(
         ctx.fail('Management key has the wrong format.')
 
 
+def _generate_random_management_key():
+    return b2a_hex(os.urandom(24)).decode()
+
+
 def _prompt_pin(ctx, prompt='Enter PIN'):
     return click.prompt(
         prompt, default='', hide_input=True, show_default=False)
 
 
+def _ensure_authenticated(
+        ctx, controller, pin=None, management_key=None,
+        require_pin_and_key=False,
+        mgm_key_prompt=None):
+
+    if controller.has_protected_key:
+        if not management_key:
+            _verify_pin(ctx, controller, pin)
+        else:
+            _authenticate(ctx, controller, management_key, mgm_key_prompt)
+    else:
+        if require_pin_and_key:
+            _verify_pin(ctx, controller, pin)
+        _authenticate(ctx, controller, management_key, mgm_key_prompt)
+
+
 def _verify_pin(ctx, controller, pin):
+    if not pin:
+        pin = _prompt_pin(ctx)
     try:
         controller.verify(pin, touch_callback=prompt_for_touch)
     except APDUError:
         ctx.fail('PIN verification failed.')
 
 
-def _authenticate(ctx, controller, management_key):
+def _authenticate(ctx, controller, management_key, mgm_key_prompt):
+    if not management_key:
+        if mgm_key_prompt is None:
+            management_key = _prompt_management_key(ctx)
+        else:
+            management_key = _prompt_management_key(ctx, mgm_key_prompt)
     try:
         controller.authenticate(management_key, touch_callback=prompt_for_touch)
     except APDUError:
