@@ -31,11 +31,14 @@ import struct
 import hashlib
 import re
 from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
-from cryptography.hazmat.primitives import hashes
+from cryptography.hazmat.primitives import hashes, serialization
 from cryptography.hazmat.backends import default_backend
-from enum import IntEnum
+from cryptography import x509
+from enum import Enum, IntEnum, unique
 from base64 import b32decode
 from binascii import b2a_hex, a2b_hex
+from OpenSSL import crypto
+
 try:
     from urlparse import urlparse, parse_qs
     from urllib import unquote
@@ -53,6 +56,16 @@ class BitflagEnum(IntEnum):
         return flags & check == check
 
 
+@unique
+class AID(bytes, Enum):
+    OTP = b'\xa0\x00\x00\x05\x27\x20\x01'
+    MGR = b'\xa0\x00\x00\x05\x27\x47\x11\x17'
+    OPGP = b'\xd2\x76\x00\x01\x24\x01'
+    OATH = b'\xa0\x00\x00\x05\x27\x21\x01'
+    PIV = b'\xa0\x00\x00\x03\x08'
+
+
+@unique
 class CAPABILITY(BitflagEnum):
     OTP = 0x01
     U2F = 0x02
@@ -67,6 +80,7 @@ class CAPABILITY(BitflagEnum):
         return CAPABILITY.OPGP | CAPABILITY.OATH | CAPABILITY.PIV
 
 
+@unique
 class TRANSPORT(BitflagEnum):
     OTP = CAPABILITY.OTP
     U2F = CAPABILITY.U2F
@@ -234,7 +248,7 @@ def derive_key(salt, passphrase):
 
 
 def format_code(code, digits=6, steam=False):
-    STEAM_CHAR_TABLE = "23456789BCDFGHJKMNPQRTVWXY"
+    STEAM_CHAR_TABLE = '23456789BCDFGHJKMNPQRTVWXY'
     if steam:
         chars = []
         for i in range(5):
@@ -302,3 +316,89 @@ def parse_b32_key(key):
     key = key.upper().replace(' ', '')
     key += '=' * (-len(key) % 8)  # Support unpadded
     return b32decode(key)
+
+
+def parse_private_key(data, password):
+    """
+    Identifies, decrypts and returns a cryptography private key object.
+    """
+    # PEM
+    if data.startswith(b'-----'):
+        if b'ENCRYPTED' in data:
+            if password is None:
+                raise TypeError('No password provided for encrypted key.')
+        try:
+            return serialization.load_pem_private_key(
+                data, password, backend=default_backend())
+        except ValueError:
+            # Cryptography raises ValueError if decryption fails.
+            raise
+        except:
+            pass
+
+    # PKCS12
+    if is_pkcs12(data):
+        try:
+            p12 = crypto.load_pkcs12(data, password)
+            data = crypto.dump_privatekey(
+                crypto.FILETYPE_PEM, p12.get_privatekey())
+            return serialization.load_pem_private_key(
+                data, password=None, backend=default_backend())
+        except crypto.Error as e:
+            raise ValueError(e)
+
+    # DER
+    try:
+        return serialization.load_der_private_key(
+            data, password, backend=default_backend())
+    except:
+        pass
+
+    # All parsing failed
+    raise ValueError('Could not parse private key.')
+
+
+def parse_certificate(data, password):
+    """
+    Identifies, decrypts and returns a cryptography x509 certficate.
+    """
+    # PEM
+    if data.startswith(b'-----'):
+        try:
+            return x509.load_pem_x509_certificate(data, default_backend())
+        except:
+            pass
+
+    # PKCS12
+    if is_pkcs12(data):
+        try:
+            p12 = crypto.load_pkcs12(data, password)
+            data = crypto.dump_certificate(
+                crypto.FILETYPE_PEM, p12.get_certificate())
+            return x509.load_pem_x509_certificate(data, default_backend())
+        except crypto.Error as e:
+            raise ValueError(e)
+
+    # DER
+    try:
+        return x509.load_der_x509_certificate(data, default_backend())
+    except:
+        pass
+
+    raise ValueError('Could not parse certificate.')
+
+
+def is_pkcs12(data):
+    """
+    Tries to identify a PKCS12 container.
+    The PFX PDU version is assumed to be v3.
+    See: https://tools.ietf.org/html/rfc7292.
+    """
+    if isinstance(data, bytes):
+        tlv = Tlv(data)
+        if tlv.tag == 0x30:
+            header = Tlv(tlv.value)
+            return header.tag == 0x02 and header.value == b'\x03'
+        return False
+    else:
+        return False
